#include "Player.hpp"
#include <Graphics/GLExtensions.hpp>
#include <Graphics/GLSLShader.hpp>
#include <Basis/Error.hpp>
#include <Vector/Box.hpp>
#include <Vector/VectorGL.hpp>

#include <algorithm>

#include <highgui.h>
#include <cv.h>

using std::swap;

REQUIRE_GL_EXTENSION( GL_ARB_pixel_buffer_object )
REQUIRE_GL_EXTENSION( GL_ARB_vertex_buffer_object )
REQUIRE_GL_EXTENSION( GL_EXT_framebuffer_object )
REQUIRE_GL_EXTENSION( GL_ARB_texture_float )
REQUIRE_GL_EXTENSION( GL_ARB_color_buffer_float )
REQUIRE_GL_EXTENSION( GL_ARB_shader_objects )
REQUIRE_GL_EXTENSION( GL_ARB_multitexture )
REQUIRE_GL_EXTENSION( GL_EXT_blend_func_separate )

Graphics::ProgramObjectRef spawn_posvel_shader = NULL;
Graphics::ProgramObjectRef spawn_color_shader = NULL;
Graphics::ProgramObjectRef update_posvel_shader = NULL;
Graphics::ProgramObjectRef update_color_shader = NULL;
Graphics::ProgramObjectRef render_field_color_shader = NULL;
Graphics::ProgramObjectRef render_field_vel_shader = NULL;
Graphics::ProgramObjectRef downsample_field_shader = NULL;
Graphics::ProgramObjectRef upsample_field_shader = NULL;
Graphics::ProgramObjectRef view_field_shader = NULL;
Graphics::ProgramObjectRef difference_shader = NULL;
Graphics::ProgramObjectRef render_instance_shader = NULL;

void alloc_rect_tex(unsigned int width, unsigned int height, GLuint &tex, GLenum format) {
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, tex);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, format, width, height, 0, GL_RGB, GL_FLOAT, NULL);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
	Graphics::gl_errors("alloc_rect_tex");
}

void alloc_2d_tex(unsigned int width, unsigned int height, GLuint &tex, GLenum format) {
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexImage2D(GL_TEXTURE_2D, 0, format, width, height, 0, GL_RGB, GL_FLOAT, NULL);
	glBindTexture(GL_TEXTURE_2D, 0);
	Graphics::gl_errors("alloc_2d_tex");
}

void alloc_fb(GLuint tex, GLuint &fb) {
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_RECTANGLE_ARB, tex, 0);

	GLenum ret = glCheckFramebufferStatusEXT( GL_FRAMEBUFFER_EXT );
	if (ret != GL_FRAMEBUFFER_COMPLETE_EXT) {
		LOG_WARNING("FRAMEBUFFER NOT COMPLETE!");
	}
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	Graphics::gl_errors("alloc_fb");

}

void alloc_2d_fb(GLuint tex, GLuint &fb) {
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex, 0);

	GLenum ret = glCheckFramebufferStatusEXT( GL_FRAMEBUFFER_EXT );
	if (ret != GL_FRAMEBUFFER_COMPLETE_EXT) {
		LOG_WARNING("FRAMEBUFFER NOT COMPLETE!");
	}
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	Graphics::gl_errors("alloc_fb");

}

TexFb::TexFb(unsigned int _width, unsigned int _height) : width(_width), height(_height) {
	alloc_2d_tex(width, height, tex, GL_RGBA16F_ARB);
	alloc_2d_fb(tex, fb);
}

Guidance::Guidance(unsigned int _width, unsigned int _height, vector< Vector3f > const &_motion) {
	width = _width;
	height = _height;
	assert(_motion.size() == width * height);
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, tex);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGB16F_ARB, width, height, 0, GL_RGB, GL_FLOAT, &_motion[0]);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
	Graphics::gl_errors("add frame");
}

Player::Player() : field_color(FieldWidth, FieldHeight), time_acc(0.0f) {

	sum_width = 32;
	sum_height = 32;

	world_size.x = FieldWidth / float(FieldHeight) * 2.0f;
	world_size.y = 2.0f;
	if (world_size.x > 2.0f * Graphics::aspectf()) {
		world_size.x = 2.0f * Graphics::aspectf();
		world_size.y = FieldHeight / float(FieldWidth) * 2.0f * Graphics::aspectf();
	}

	assert(world_size.x <= 2.0f * Graphics::aspectf());
	assert(world_size.y <= 2.0f);

	spawn_point = 0;
	//various particle buffers:
	alloc_rect_tex(ParticlesWidth, ParticlesHeight, posvel_tex, GL_RGBA32F_ARB);
	alloc_rect_tex(ParticlesWidth, ParticlesHeight, posvel_temp_tex, GL_RGBA32F_ARB);
	alloc_rect_tex(ParticlesWidth, ParticlesHeight, color_tex, GL_RGB16F_ARB);
	alloc_rect_tex(ParticlesWidth, ParticlesHeight, color_temp_tex, GL_RGB16F_ARB);
	alloc_fb(posvel_tex, posvel_fb);
	alloc_fb(posvel_temp_tex, posvel_temp_fb);
	alloc_fb(color_tex, color_fb);
	alloc_fb(color_temp_tex, color_temp_fb);
	{
		glGenTextures(1, &spawn_tex);
		glBindTexture(GL_TEXTURE_2D, spawn_tex);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		vector< Vector3f > pos;
		for (unsigned int x = 0; x < FieldWidth; ++x) {
			for (unsigned int y = 0; y < FieldHeight; ++y) {
				pos.push_back(make_vector< float >(x / float(FieldWidth) * world_size.x, y / float(FieldHeight) * world_size.y, 0.0f));
				swap(pos[rand() % pos.size()], pos.back());
			}
		}
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB16F_ARB, FieldWidth, FieldHeight, 0, GL_RGB, GL_FLOAT, &pos[0]);
		glBindTexture(GL_TEXTURE_2D, 0);
		Graphics::gl_errors("alloc_2d_tex -- for spawn");
	}
	//particle rendering vbo's:
	glGenBuffersARB(1, &posvel_vbo);
	glGenBuffersARB(1, &color_vbo);
	glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB, posvel_vbo);
	glBufferDataARB(GL_PIXEL_PACK_BUFFER_ARB, ParticlesWidth * ParticlesHeight * sizeof(float) * 4, NULL, GL_DYNAMIC_DRAW);
	glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB, 0);
	glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB, color_vbo);
	glBufferDataARB(GL_PIXEL_PACK_BUFFER_ARB, ParticlesWidth * ParticlesHeight * sizeof(float) * 4, NULL, GL_DYNAMIC_DRAW);
	glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB, 0);
	//field:
	field.push_back(TexFb(FieldWidth, FieldHeight));
	while (field.back().width > 1 && field.back().height > 1) {
		field.push_back(TexFb(field.back().width / 2, field.back().height / 2));
	}
	field_temp.push_back(TexFb(FieldWidth, FieldHeight));
	while (field_temp.back().width > 1 && field_temp.back().height > 1) {
		field_temp.push_back(TexFb(field_temp.back().width / 2, field_temp.back().height / 2));
	}
	assert(field_temp.size() == field.size());
	//textures for sum (e.g. check vs field):
	alloc_rect_tex(sum_width, sum_height, sum_tex, GL_RGBA32F_ARB);
	alloc_rect_tex(sum_width, sum_height, sum_temp_tex, GL_RGBA32F_ARB);
	alloc_fb(sum_tex, sum_fb);
	alloc_fb(sum_temp_tex, sum_temp_fb);
	//set clamping "right":
	glClampColorARB(GL_CLAMP_VERTEX_COLOR_ARB, GL_FALSE);
	glClampColorARB(GL_CLAMP_FRAGMENT_COLOR_ARB, GL_FALSE);
	glClampColorARB(GL_CLAMP_READ_COLOR_ARB, GL_FALSE);
	//load update shaders:
	{
		Graphics::ShaderObjectRef frag = Graphics::get_shader_object("shaders/spawn_posvel.glsl", GL_FRAGMENT_SHADER_ARB);
		assert(frag.ref);
		spawn_posvel_shader = Graphics::get_program_object(frag);
		assert(spawn_posvel_shader.ref);
		glUseProgramObjectARB(spawn_posvel_shader->handle);
		glUniform1iARB(glGetUniformLocationARB(spawn_posvel_shader->handle, "pos_tex"), 0);
		glUniform1iARB(glGetUniformLocationARB(spawn_posvel_shader->handle, "vel_tex"), 1);
		glUniform2fARB(glGetUniformLocationARB(spawn_posvel_shader->handle, "world"), world_size.x, world_size.y);
		glUseProgramObjectARB(0);
	}
	{
		Graphics::ShaderObjectRef frag = Graphics::get_shader_object("shaders/spawn_color.glsl", GL_FRAGMENT_SHADER_ARB);
		assert(frag.ref);
		spawn_color_shader = Graphics::get_program_object(frag);
		assert(spawn_color_shader.ref);
		glUseProgramObjectARB(spawn_color_shader->handle);
		glUniform1iARB(glGetUniformLocationARB(spawn_color_shader->handle, "pos_tex"), 0);
		glUniform1iARB(glGetUniformLocationARB(spawn_color_shader->handle, "color_tex"), 1);
		glUniform2fARB(glGetUniformLocationARB(spawn_color_shader->handle, "world"), world_size.x, world_size.y);
		glUseProgramObjectARB(0);
	}
	{
		Graphics::ShaderObjectRef frag = Graphics::get_shader_object("shaders/update_posvel.glsl", GL_FRAGMENT_SHADER_ARB);
		assert(frag.ref);
		update_posvel_shader = Graphics::get_program_object(frag);
		assert(update_posvel_shader.ref);
		glUseProgramObjectARB(update_posvel_shader->handle);
		glUniform1iARB(glGetUniformLocationARB(update_posvel_shader->handle, "posvel_tex"), 0);
		glUniform1iARB(glGetUniformLocationARB(update_posvel_shader->handle, "vel_tex"), 1);
		glUniform2fARB(glGetUniformLocationARB(update_posvel_shader->handle, "world"), world_size.x, world_size.y);
		glUseProgramObjectARB(0);
	}
	{
		Graphics::ShaderObjectRef frag = Graphics::get_shader_object("shaders/update_color.glsl", GL_FRAGMENT_SHADER_ARB);
		assert(frag.ref);
		update_color_shader = Graphics::get_program_object(frag);
		assert(update_color_shader.ref);
		glUseProgramObjectARB(update_color_shader->handle);
		glUniform1iARB(glGetUniformLocationARB(update_color_shader->handle, "posvel_tex"), 0);
		glUniform1iARB(glGetUniformLocationARB(update_color_shader->handle, "color_tex"), 1);
		glUniform1iARB(glGetUniformLocationARB(update_color_shader->handle, "field_tex"), 2);
		glUniform1iARB(glGetUniformLocationARB(update_color_shader->handle, "cam_tex"), 3);
		glUniform2fARB(glGetUniformLocationARB(update_color_shader->handle, "world"), world_size.x, world_size.y);
		glUseProgramObjectARB(0);
	}
	{
		Graphics::ShaderObjectRef vert = Graphics::get_shader_object("shaders/render_field_color.glsl", GL_VERTEX_SHADER_ARB);
		assert(vert.ref);
		render_field_color_shader = Graphics::get_program_object(vert);
		assert(render_field_color_shader.ref);
		glUseProgramObjectARB(render_field_color_shader->handle);
		glUniform2fARB(glGetUniformLocationARB(render_field_color_shader->handle, "world"), world_size.x, world_size.y);
		glUseProgramObjectARB(0);
	}
	{
		Graphics::ShaderObjectRef vert = Graphics::get_shader_object("shaders/render_field_vel.glsl", GL_VERTEX_SHADER_ARB);
		assert(vert.ref);
		render_field_vel_shader = Graphics::get_program_object(vert);
		assert(render_field_vel_shader.ref);
		glUseProgramObjectARB(render_field_vel_shader->handle);
		glUniform2fARB(glGetUniformLocationARB(render_field_vel_shader->handle, "world"), world_size.x, world_size.y);
		glUseProgramObjectARB(0);
	}
	{
		Graphics::ShaderObjectRef vert = Graphics::get_shader_object("shaders/render_instance.glsl", GL_FRAGMENT_SHADER_ARB);
		assert(vert.ref);
		render_instance_shader = Graphics::get_program_object(vert);
		assert(render_instance_shader.ref);
		glUseProgramObjectARB(render_instance_shader->handle);
		glUniform1iARB(glGetUniformLocationARB(render_instance_shader->handle, "instance_tex"), 0);
		glUseProgramObjectARB(0);
	}
	{
		Graphics::ShaderObjectRef frag = Graphics::get_shader_object("shaders/downsample_field.glsl", GL_FRAGMENT_SHADER_ARB);
		assert(frag.ref);
		downsample_field_shader = Graphics::get_program_object(frag);
		assert(downsample_field_shader.ref);
		glUseProgramObjectARB(downsample_field_shader->handle);
		glUniform1iARB(glGetUniformLocationARB(downsample_field_shader->handle, "large_tex"), 0);
		glUseProgramObjectARB(0);
	}
	{
		Graphics::ShaderObjectRef frag = Graphics::get_shader_object("shaders/upsample_field.glsl", GL_FRAGMENT_SHADER_ARB);
		assert(frag.ref);
		upsample_field_shader = Graphics::get_program_object(frag);
		assert(upsample_field_shader.ref);
		glUseProgramObjectARB(upsample_field_shader->handle);
		glUniform1iARB(glGetUniformLocationARB(upsample_field_shader->handle, "large_tex"), 0);
		glUniform1iARB(glGetUniformLocationARB(upsample_field_shader->handle, "small_tex"), 1);
		glUseProgramObjectARB(0);
	}
	{
		Graphics::ShaderObjectRef frag = Graphics::get_shader_object("shaders/view_field.glsl", GL_FRAGMENT_SHADER_ARB);
		assert(frag.ref);
		view_field_shader = Graphics::get_program_object(frag);
		assert(view_field_shader.ref);
		glUseProgramObjectARB(view_field_shader->handle);
		glUniform1iARB(glGetUniformLocationARB(view_field_shader->handle, "field_tex"), 0);
		glUseProgramObjectARB(0);
	}
	{
		Graphics::ShaderObjectRef frag = Graphics::get_shader_object("shaders/difference.glsl", GL_FRAGMENT_SHADER_ARB);
		assert(frag.ref);
		difference_shader = Graphics::get_program_object(frag);
		assert(difference_shader.ref);
		glUseProgramObjectARB(difference_shader->handle);
		glUniform1iARB(glGetUniformLocationARB(difference_shader->handle, "instance_tex"), 1);
		glUniform1iARB(glGetUniformLocationARB(difference_shader->handle, "field_tex"), 0);
		glUseProgramObjectARB(0);
	}
	//Some guidance, perhaps:

	{
		const unsigned int Width = 32;
		const unsigned int Height = 29;
		vector< Vector3f > motion(Width * Height);
		for (unsigned int x = 0; x < Width; ++x) {
			for (unsigned int y = 0; y < Height; ++y) {
				motion[y * Width + x] = make_vector(0.1f, 0.0f, 1.0f);
			}
		}
		guidance.push_back(Guidance(Width, Height, motion));
	}

	{
		const unsigned int Width = 16;
		const unsigned int Height = 16;
		vector< Vector3f > motion(Width * Height);
		for (unsigned int x = 0; x < Width; ++x) {
			for (unsigned int y = 0; y < Height; ++y) {
				Vector2f pos;
				pos.x = (x / float(Width - 1) - 0.5f) * 2.0f;
				pos.y = (y / float(Height - 1) - 0.5f) * 2.0f * float(Height) / float(Width);
				float str = 1.0f - length(pos);
				if (str < 0.0f) str = 0.0f;
				motion[y * Width + x] = make_vector(normalize(pos) * 0.1f, str);
			}
		}
		guidance.push_back(Guidance(Width, Height, motion));
	}
	{
		const unsigned int Width = 32;
		const unsigned int Height = 32;
		vector< Vector3f > motion(Width * Height);
		for (unsigned int x = 0; x < Width; ++x) {
			for (unsigned int y = 0; y < Height; ++y) {
				Vector2f pos;
				pos.x = (x / float(Width - 1) - 0.5f) * 2.0f;
				pos.y = (y / float(Height - 1) - 0.5f) * 2.0f * float(Height) / float(Width);
				float str = 1.0f - length(pos);
				if (str < 0.0f) str = 0.0f;
				motion[y * Width + x] = make_vector(perpendicular(pos) * 0.5f - 0.5f * pos, str);
			}
		}
		guidance.push_back(Guidance(Width, Height, motion));
	}

	mouse_pos = world_size * 0.5f;
	mouse_rot = 0.0f;
	mouse_scale = 0.6f;

	show_boxes = false;
	
	cam_image = 0;
	cam_weight = 0.0f;
	cam_color = 0;
	cam_width = 1;
	cam_height = 1;
	cam_color_width = 1;
	cam_color_height = 1;

	init();
}

Player::~Player() {
}

void Player::handle_event(SDL_Event const &event) {
	if (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_SPACE) {
		init();
		show_boxes = !show_boxes;
	}
	if (event.type == SDL_MOUSEMOTION) {
		mouse_pos.x = (event.motion.x / float(Graphics::screen_x) - 0.5f) * 2.0f * Graphics::aspectf() / 0.9f;
		mouse_pos.y = (event.motion.y / float(Graphics::screen_y) - 0.5f) *-2.0f / 0.9f;
		mouse_pos += world_size * 0.5f;
	}
	if (event.type == SDL_MOUSEBUTTONDOWN && event.button.button == SDL_BUTTON_WHEELUP) {
		mouse_rot += 0.1f * M_PI;
	}
	if (event.type == SDL_MOUSEBUTTONDOWN && event.button.button == SDL_BUTTON_WHEELDOWN) {
		mouse_rot -= 0.1f * M_PI;
	}
	if (event.type == SDL_MOUSEBUTTONDOWN && event.button.button == SDL_BUTTON_LEFT) {
		instances.push_back(Instance());
		instances.back().source = rand() % guidance.size();
		instances.back().pos = mouse_pos;
		instances.back().rot = mouse_rot;
		instances.back().scale = mouse_scale;
		instances.back().lifetime = 0.0f;
	}
}

extern CvCapture *capture;

void Player::new_frame() {
	if (!capture) return;
	IplImage *frame = cvQueryFrame( capture );
	if (!frame) {
		//quit = true;
		return;
	}
	if (frame->nChannels != 3) {
		LOG_ERROR("Frame has unexpected channels:" << frame->nChannels);
		return;
	}
	if (frame->depth != IPL_DEPTH_8U) {
		LOG_ERROR("Frame has unexpected depth: " << frame->depth);
		return;
	}
	if (frame->dataOrder != 0) {
		LOG_ERROR("Frame isn't interleaved.");
		return;
	}
	cam_width = frame->width / 4;
	cam_height = frame->height / 4;
	static IplImage *grey = cvCreateImage(cvSize(frame->width, frame->height), IPL_DEPTH_8U, 1);
	cvCvtColor(frame, grey, CV_RGB2GRAY);
	static IplImage *a = cvCreateImage(cvSize(cam_width, cam_height), IPL_DEPTH_8U, 1);
	static IplImage *b = cvCreateImage(cvSize(cam_width, cam_height), IPL_DEPTH_8U, 1);
	static IplImage *velx = cvCreateImage(cvSize(cam_width, cam_height), IPL_DEPTH_32F, 1);
	static IplImage *vely = cvCreateImage(cvSize(cam_width, cam_height), IPL_DEPTH_32F, 1);
	//Assumption being that all frames are the same size:
	assert((unsigned)a->width == cam_width);
	assert((unsigned)a->height == cam_height);

	swap(a,b);
	cvResize(grey, a);

	static bool have_b = false;
	if (!have_b) {
		have_b = true;
		return;
	}

	cvCalcOpticalFlowHS( b, a, 1, velx, vely, 1.0f, cvTermCriteria(CV_TERMCRIT_ITER, 10, 0.1f));
	assert((unsigned)velx->widthStep == cam_width * sizeof(float));
	static vector< Vector3f > tex_src(cam_width*cam_height);
	float inv = 1.0f / 10.0f; //float(cam_width);
	for (unsigned int i = 0; i < tex_src.size(); ++i) {
		tex_src[i].x = ((float*)velx->imageData)[i] * inv;
		//((uint8_t *)a->imageData)[i] / 255.0f;
		tex_src[i].y = ((float*)vely->imageData)[i] * inv;
		tex_src[i].z = 1.0f;
	}
	if (cam_image == 0) {
		glGenTextures(1, &cam_image);
	}
	cam_weight = 1.0f;
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, cam_image);
	glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGB16F_ARB, cam_width, cam_height, 0, GL_RGB, GL_FLOAT, &tex_src[0]);

	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
	if (cam_color == 0) {
		glGenTextures(1, &cam_color);
	}
	cam_color_width = frame->width;
	cam_color_height = frame->height;
	glBindTexture(GL_TEXTURE_2D, cam_color);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, cam_color_width, cam_color_height, 0, GL_BGR, GL_UNSIGNED_BYTE, frame->imageData);
	glBindTexture(GL_TEXTURE_2D, 0);


}

void Player::update(float elapsed_time) {
	static bool skip = false;
	if (!skip) {
		new_frame();
		skip = true;
	} else {
		skip = false;
	}

	time_acc += elapsed_time;
	while (time_acc >= 0.0f) {
		time_acc -= Tick;
		tick();
	}
}

void Player::draw() {
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glScalef(0.9f / Graphics::aspectf(), 0.9f, 1.0f);
	glTranslatef(-0.5f * world_size.x, -0.5f * world_size.y, 0.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	//draw field color:
	glDisable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, field_color.tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(0.0f, 0.0f);
	glTexCoord2f(1.0f, 0.0f);
	glVertex2f(world_size.x, 0.0f);
	glTexCoord2f(1.0f, 1.0f);
	glVertex2f(world_size.x, world_size.y);
	glTexCoord2f(0.0f, 1.0f);
	glVertex2f(0.0f, world_size.y);
	glEnd();
	glBindTexture(GL_TEXTURE_2D, 0);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);

	glPointSize(5.0f);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	//render vbo's:
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, posvel_vbo);
	glVertexPointer(2, GL_FLOAT, sizeof(float)*4, 0);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, color_vbo);
	glColorPointer(4, GL_FLOAT, sizeof(float)*4, 0);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glDrawArrays(GL_POINTS, 0, ParticlesWidth * ParticlesHeight);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	Graphics::gl_errors("Render vbo");


	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	/*
	//draw field:
	glBindTexture(GL_TEXTURE_2D, field[0].tex);
	glUseProgramObjectARB(view_field_shader->handle);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(0.0f, 0.0f);
	glTexCoord2f(1.0f, 0.0f);
	glVertex2f(world_size.x, 0.0f);
	glTexCoord2f(1.0f, 1.0f);
	glVertex2f(world_size.x, world_size.y);
	glTexCoord2f(0.0f, 1.0f);
	glVertex2f(0.0f, world_size.y);
	glEnd();
	glBindTexture(GL_TEXTURE_2D, 0);
	glUseProgramObjectARB(0);
*/
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);


	//show instances:
	if (show_boxes) {
		for (vector< Instance >::const_iterator i = instances.begin(); i != instances.end(); ++i) {
			Vector2f dir = make_vector(cosf(i->rot), sinf(i->rot));
			Vector2f perp = perpendicular(dir);
			dir *= 0.5f * i->scale;
			perp *= 0.5f * i->scale / guidance[i->source].width * guidance[i->source].height;
			glBegin(GL_LINE_LOOP);
			glColor3f(0.0f, 0.0f, 1.0f);
			glVertex(i->pos-dir-perp);
			glVertex(i->pos+dir-perp);
			glVertex(i->pos+dir+perp);
			glVertex(i->pos-dir+perp);
			glEnd();
		}
	}
	if (show_boxes) {
		Vector2f dir = make_vector(cosf(mouse_rot), sinf(mouse_rot));
		Vector2f perp = perpendicular(dir);
		dir *= 0.5f * mouse_scale;
		perp *= 0.5f * mouse_scale;
		glBegin(GL_LINE_LOOP);
		glColor3f(0.0f, 0.0f, 1.0f);
		glVertex(mouse_pos-dir-perp);
		glVertex(mouse_pos+dir-perp);
		glVertex(mouse_pos+dir+perp);
		glVertex(mouse_pos-dir+perp);
		glEnd();

/*
		//show sum texture:
		glEnable(GL_TEXTURE_RECTANGLE_ARB);
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, sum_tex);
		glDisable(GL_BLEND);
		glBegin(GL_QUADS);
		glColor3f(1.0f, 1.0f, 1.0f);
		glTexCoord2f(0,0);
		glVertex(mouse_pos-dir-perp);
		glTexCoord2f(sum_width,0);
		glVertex(mouse_pos+dir-perp);
		glTexCoord2f(sum_width,sum_height);
		glVertex(mouse_pos+dir+perp);
		glTexCoord2f(0,sum_height);
		glVertex(mouse_pos-dir+perp);
		glEnd();
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
		glDisable(GL_TEXTURE_RECTANGLE_ARB);
*/
	}

/*
	if (cam_image) {
		glEnable(GL_TEXTURE_RECTANGLE_ARB);
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, cam_image);
		glColor3f(1.0f, 1.0f, 1.0f);
		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex2f(0.0f, world_size.y);
		glTexCoord2f(cam_width,0);
		glVertex2f(1.0f, world_size.y);
		glTexCoord2f(cam_width,cam_height);
		glVertex2f(1.0f, world_size.y * 1.5);
		glTexCoord2f(0,cam_height);
		glVertex2f(0.0f, world_size.y * 1.5);
		glEnd();
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
		glDisable(GL_TEXTURE_RECTANGLE_ARB);
	}
*/
	Graphics::gl_errors("draw field");
}

void Player::init() {
	glDisable(GL_BLEND);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0,0,ParticlesWidth,ParticlesHeight);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, posvel_temp_fb);

	glBegin(GL_QUADS);
	glColor4f(0.0f,0.0f, 0.0f, 0.0f);
	glVertex2f(-1.0f,-1.0f);
	glColor4f(world_size.x,0.0f, 0.0f, 0.0f);
	glVertex2f(1.0f,-1.0f);
	glColor4f(world_size.x, world_size.y, 0.0f, 0.0f);
	glVertex2f(1.0f, 1.0f);
	glColor4f(0.0f, world_size.y, 0.0f, 0.0f);
	glVertex2f(-1.0f,1.0f);
	glEnd();

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, color_temp_fb);
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	glPopAttrib();
	swap(posvel_tex, posvel_temp_tex);
	swap(posvel_fb, posvel_temp_fb);
	swap(color_tex, color_temp_tex);
	swap(color_fb, color_temp_fb);


	tick(); //to make sure it gets read.
}

void Player::check_instance(Vector2f pos, float rot, float scale_width, unsigned int width, unsigned int height, GLuint tex, float &sum, Vector2f &vel, float &rot_vel) {
	//Ok, we want the sum-square difference, at optimal rotation and velocity,
	//between our example patch (width x height, in 'tex') and the field.
	//What we will do is compute this sum parametrically (per element) into
	//sum_tex, then perform a reduction and read the parameters back,
	//then solve for the optimal vel, rot_vel and return the sum.
	
	glDisable(GL_BLEND);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glPushAttrib(GL_VIEWPORT_BIT);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, sum_fb);
	glViewport(0,0,sum_width,sum_height);
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT);

	//Challenge here: set up texture coordinates properly for the two sources,
	//also rasterize proper width/height...
	Vector2f render_size;
	{ //figure out a good render_size:
		render_size.x = (float)sum_width;
		render_size.y = render_size.x / width * height;
		if (render_size.y > sum_width) {
			render_size.y = sum_width;
			render_size.x = render_size.y / height * width;
		}
		assert(render_size.x <= sum_width);
		assert(render_size.y <= sum_height);
		//put into the standard [-1,1] box:
		render_size.x /= sum_width;
		render_size.y /= sum_height;
		render_size *= 2.0f;
	}

	Vector2f dir = make_vector(cosf(rot), sinf(rot));
	Vector2f perp = perpendicular(dir);

	dir *= 0.5f * scale_width;
	perp *= 0.5f * scale_width / width * height;

	//store these world-space directions out:
	Vector2f rdir = dir;
	Vector2f rperp = perp;

	pos.x /= world_size.x;
	pos.y /= world_size.y;
	dir.x /= world_size.x;
	dir.y /= world_size.y;
	perp.x /= world_size.x;
	perp.y /= world_size.y;

	glUseProgramObjectARB(difference_shader->handle);
	glUniform2fARB(glGetUniformLocationARB(difference_shader->handle, "dir"), cosf(rot), sinf(rot));

	//this might be about right:
	glActiveTextureARB(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, tex);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glActiveTextureARB(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, field[0].tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glBegin(GL_QUADS);
	glMultiTexCoord2fARB(GL_TEXTURE1, 0, 0);
	glMultiTexCoord2fARB(GL_TEXTURE2, -rdir.x-rperp.x, -rdir.y-rperp.y);
	glTexCoord(pos - dir - perp);
	glVertex2f(-1.0f,-1.0f);
	glMultiTexCoord2fARB(GL_TEXTURE1, width, 0);
	glMultiTexCoord2fARB(GL_TEXTURE2,  rdir.x-rperp.x,  rdir.y-rperp.y);
	glTexCoord(pos + dir - perp);
	glVertex2f(-1.0f + render_size.x,-1.0f);
	glMultiTexCoord2fARB(GL_TEXTURE1, width, height);
	glMultiTexCoord2fARB(GL_TEXTURE2,  rdir.x+rperp.x,  rdir.y+rperp.y);
	glTexCoord(pos + dir + perp);
	glVertex2f(-1.0f + render_size.x,-1.0f + render_size.y);
	glMultiTexCoord2fARB(GL_TEXTURE1, 0, height);
	glMultiTexCoord2fARB(GL_TEXTURE2, -rdir.x+rperp.x, -rdir.y+rperp.y);
	glTexCoord(pos - dir + perp);
	glVertex2f(-1.0f,-1.0f + render_size.y);
	glEnd();
	glActiveTextureARB(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
	glActiveTextureARB(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, 0);

	glUseProgramObjectARB(0);

	vector< Vector4f > buffer(sum_width*sum_height);
	glReadPixels(0,0,sum_width,sum_height,GL_RGBA, GL_FLOAT, &buffer[0]);

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	glPopAttrib();
	
	Vector4f acc = make_vector(0.0f, 0.0f, 0.0f, 0.0f);

	for (unsigned int i = 0; i < buffer.size(); ++i) {
		acc += buffer[i];
	}
	acc /= buffer.size();

	//LOG_INFO("Acc: " << acc);

	//Partial w.r.t trans.x:
	vel = make_vector(-acc.z,-acc.w);
	//Partial w.r.t. rot:
	//2*r*(rot*rot) + (rot*scale)
	rot_vel = -acc.x * 100.0f; //this is a bit of a fudged thing.
	sum = vel * vel + vel * make_vector( acc.z, acc.w) + acc.y;
}

void Player::tick() {
	//Basic idea:
	//(a) particles write velocity and color into field.
	//(b) guidance writes velocity and color into field (alpha blended).
	//(c) some sort of compressability correction?
	//(d) particles read (some percentage) velocity and color from field.
	//     (weighted by alpha, perhaps?)
	//(e) particles respawned to low-density areas.

	
	{ //read positions into vbo:
		glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB, posvel_vbo);
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, posvel_tex);
		glGetTexImage(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA, GL_FLOAT, 0);
		glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB, color_vbo);
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, color_tex);
		glGetTexImage(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA, GL_FLOAT, 0);
		glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB, 0);
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
	
		Graphics::gl_errors("Copy to vbo");
	}

	glPushAttrib(GL_VIEWPORT_BIT);

	//we'll actually do two interpolation passes with field; first will
	//be color, second velocity:
	for (unsigned int field_iter = 0; field_iter < 2; ++field_iter) {
	//render to field:
	{
		glViewport(0,0,FieldWidth,FieldHeight);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, field[0].fb);
		glEnable(GL_BLEND);
		glBlendFuncSeparateEXT(GL_SRC_ALPHA, GL_ONE, GL_ONE, GL_ONE);
		glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		glClear(GL_COLOR_BUFFER_BIT);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glTranslatef(-1.0f, -1.0f, 0.0f);
		glScalef(2.0f / world_size.x, 2.0f / world_size.y, 1.0f);

		glPointSize(1.0f);
		//render vbo's:
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, posvel_vbo);
		glVertexPointer(4, GL_FLOAT, sizeof(float)*4, 0);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, color_vbo);
		glColorPointer(4, GL_FLOAT, sizeof(float)*4, 0);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_COLOR_ARRAY);
		if (field_iter == 0) {
			glUseProgramObjectARB(render_field_color_shader->handle);
		} else {
			glUseProgramObjectARB(render_field_vel_shader->handle);
		}
		glDrawArrays(GL_POINTS, 0, ParticlesWidth * ParticlesHeight);
		glUseProgramObjectARB(0);
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_COLOR_ARRAY);

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
		Graphics::gl_errors("render to field");
	}
	//downsample field:
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glDisable(GL_BLEND);
	for (unsigned int i = 1; i < field.size(); ++i) {
		glViewport(0,0,field[i].width,field[i].height);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, field[i].fb);
		glBindTexture(GL_TEXTURE_2D, field[i-1].tex);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glUseProgramObjectARB(downsample_field_shader->handle);
		glBegin(GL_QUADS);
		glTexCoord2f(0.0f, 0.0f);
		glVertex2f(-1.0f,-1.0f);
		glTexCoord2f(1.0f, 0.0f);
		glVertex2f(1.0f,-1.0f);
		glTexCoord2f(1.0f, 1.0f);
		glVertex2f(1.0f, 1.0f);
		glTexCoord2f(0.0f, 1.0f);
		glVertex2f(-1.0f, 1.0f);
		glEnd();
		glUseProgramObjectARB(0);
		glBindTexture(GL_TEXTURE_2D, 0);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	}
	//upsample field:
	for (unsigned int i = field.size()-2; i < field.size(); --i) {
		glViewport(0,0,field[i].width,field[i].height);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, field_temp[i].fb);

		glActiveTextureARB(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, field[i+1].tex);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glActiveTextureARB(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, field[i].tex);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glUseProgramObjectARB(upsample_field_shader->handle);
		glBegin(GL_QUADS);
		glTexCoord2f(0.0f, 0.0f);
		glVertex2f(-1.0f,-1.0f);
		glTexCoord2f(1.0f, 0.0f);
		glVertex2f(1.0f,-1.0f);
		glTexCoord2f(1.0f, 1.0f);
		glVertex2f(1.0f, 1.0f);
		glTexCoord2f(0.0f, 1.0f);
		glVertex2f(-1.0f, 1.0f);
		glEnd();
		glUseProgramObjectARB(0);
		glActiveTextureARB(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, 0);
		glActiveTextureARB(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, 0);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
		field_temp[i].swap(field[i]);
	}

	if (field_iter == 0) {
		field_color.swap(field[0]); //save off color.
	}

	} //end for(field_iter);
	//Now we want to see about moving some guidance:
	for (vector< Instance >::iterator i = instances.begin(); i != instances.end();) {
		i->lifetime += Tick;
		if (i->lifetime > MaxLifetime) {
			i = instances.erase(i);
			continue;
		}
		float sum = 0.0f;
		Vector2f vel = make_vector(0.0f, 0.0f);
		float rot_vel = 0.0f;
		check_instance(i->pos, i->rot, i->scale, guidance[i->source].width, guidance[i->source].height, guidance[i->source].tex, sum, vel, rot_vel);
		
		i->pos += vel * Tick;
		i->rot += rot_vel * Tick;

		if (i->pos.x >= world_size.x) {
			i->pos.x = fmodf(i->pos.x, world_size.x);
		}
		if (i->pos.x < 0) {
			i->pos.x = fmodf(i->pos.x, world_size.x) + world_size.x;
		}

		if (i->pos.y >= world_size.y) {
			i->pos.y = fmodf(i->pos.y, world_size.y);
		}
		if (i->pos.y < 0) {
			i->pos.y = fmodf(i->pos.y, world_size.y) + world_size.y;
		}
		++i;
	}


	if (instances.size() < 5) {
		Instance r;
		r.pos.x = rand() / float(RAND_MAX) * world_size.x;
		r.pos.y = rand() / float(RAND_MAX) * world_size.y;
		r.rot = rand() / float(RAND_MAX) * M_PI * 2.0f;
		r.scale = rand() / float(RAND_MAX) * 0.3f + 0.3f;
		r.source = rand() % guidance.size();
		r.lifetime = 0.0f;
		float sum = 0.0f;
		Vector2f vel = make_vector(0.0f, 0.0f);
		float rot_vel = 0.0f;
		check_instance(r.pos, r.rot, r.scale, guidance[r.source].width, guidance[r.source].height, guidance[r.source].tex, sum, vel, rot_vel);

		const float Decay = 0.5f;
		const float Norm = 1.0f / (1.0f - Decay);
		static float running_avg = 0.0f;
		running_avg = running_avg * 0.5f + sum;

		static unsigned int avoid_ticks = 0;
		if (avoid_ticks > 0) {
			--avoid_ticks;
		} else if (sum  * Norm * 1.5f > running_avg) {
			instances.push_back(r);
			LOG_INFO("Spawn with " << sum << " versus " << running_avg / Norm);
			avoid_ticks = 100; //wait a whole second.
		}
	}

	//now draw instances:
	glViewport(0,0,field[0].width,field[0].height);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, field[0].fb);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glUseProgramObjectARB(render_instance_shader->handle);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	if (cam_image) { //HACK: draw motion:
		glUniform2fARB(glGetUniformLocationARB(render_instance_shader->handle, "dir"), 1.0f, 0.0f);
		glUniform1fARB(glGetUniformLocationARB(render_instance_shader->handle, "att"), 0.5f * cam_weight);
		cam_weight *= 0.9f;

		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, cam_image);
		glBegin(GL_QUADS);
		glTexCoord2f(0.0f, 0.0f);
		glVertex2f(-1.0f,-1.0f);
		glTexCoord2f(cam_width, 0.0f);
		glVertex2f( 1.0f,-1.0f);
		glTexCoord2f(cam_width, cam_height);
		glVertex2f( 1.0f, 1.0f);
		glTexCoord2f(0.0f, cam_height);
		glVertex2f(-1.0f, 1.0f);
		glEnd();
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
	} //end HACK
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	for (vector< Instance >::iterator i = instances.begin(); i != instances.end(); ++i) {
		Vector2f pos = i->pos;
		Vector2f dir = make_vector(cosf(i->rot), sinf(i->rot));
		Vector2f perp = perpendicular(dir);

		dir *= 0.5f * i->scale;
		perp *= 0.5f * i->scale / guidance[i->source].width * guidance[i->source].height;

		pos.x /= world_size.x;
		pos.y /= world_size.y;
		dir.x /= world_size.x;
		dir.y /= world_size.y;
		perp.x /= world_size.x;
		perp.y /= world_size.y;

		pos = (pos - make_vector(0.5f, 0.5f)) * 2.0f;
		dir *= 2.0f;
		perp *= 2.0f;

		vector< Vector2f > offsets;
		offsets.push_back(make_vector(0.0f, 0.0f));
		if (pos.x - fabs(dir.x) - fabs(perp.x) < -1.0f) {
			offsets.push_back(make_vector(2.0f, 0.0f));
		}
		if (pos.x + fabs(dir.x) + fabs(perp.x) >  1.0f) {
			offsets.push_back(make_vector(-2.0f, 0.0f));
		}
		unsigned int x_offsets = offsets.size();
		if (pos.y - fabs(dir.y) - fabs(perp.y) < -1.0f) {
			for (unsigned int o = 0; o < x_offsets; ++o) {
				offsets.push_back(offsets[o] + make_vector( 0.0f, 2.0f));
			}
		}
		if (pos.y + fabs(dir.y) + fabs(perp.y) >  1.0f) {
			for (unsigned int o = 0; o < x_offsets; ++o) {
				offsets.push_back(offsets[o] + make_vector( 0.0f,-2.0f));
			}
		}
		glUniform2fARB(glGetUniformLocationARB(render_instance_shader->handle, "dir"), cosf(i->rot), sinf(i->rot));
		float fade = 1.0f;
		if (i->lifetime < FadeIn) {
			fade = i->lifetime / FadeIn;
		} else if (i->lifetime > MaxLifetime - FadeOut) {
			fade = (MaxLifetime - i->lifetime) / FadeOut;
		}
		glUniform1fARB(glGetUniformLocationARB(render_instance_shader->handle, "att"), 0.1 * fade);

		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, guidance[i->source].tex);
		glBegin(GL_QUADS);
		for (vector< Vector2f >::iterator o = offsets.begin(); o != offsets.end(); ++o) {
			glTexCoord2f(0.0f, 0.0f);
			glVertex(*o+pos-dir-perp);
			glTexCoord2f(guidance[i->source].width, 0.0f);
			glVertex(*o+pos+dir-perp);
			glTexCoord2f(guidance[i->source].width, guidance[i->source].height);
			glVertex(*o+pos+dir+perp);
			glTexCoord2f(0.0f, guidance[i->source].height);
			glVertex(*o+pos-dir+perp);
		}
		glEnd();
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);

	}
	glUseProgramObjectARB(0);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	//setup spawn:
	assert((spawn_point % ParticlesWidth) + SpawnSize <= ParticlesWidth);
	Box2f spawn_box;
	spawn_box.min.x = (spawn_point % ParticlesWidth) / float(ParticlesWidth) * 2.0f - 1.0f;
	spawn_box.min.y = (spawn_point / ParticlesWidth) / float(ParticlesHeight) * 2.0f - 1.0f;
	spawn_box.max.x = ((spawn_point % ParticlesWidth) + SpawnSize) / float(ParticlesWidth) * 2.0f - 1.0f;
	spawn_box.max.y = (spawn_point / ParticlesWidth + 1) / float(ParticlesHeight) * 2.0f - 1.0f;


	//setup for particles:
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glDisable(GL_BLEND);
	glViewport(0,0,ParticlesWidth,ParticlesHeight);

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, posvel_temp_fb);

	glUseProgramObjectARB(update_posvel_shader->handle);
	glActiveTextureARB(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, field[0].tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glActiveTextureARB(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, posvel_tex);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex2f(-1.0f,-1.0f);
	glTexCoord2f(ParticlesWidth,0);
	glVertex2f(1.0f,-1.0f);
	glTexCoord2f(ParticlesWidth,ParticlesHeight);
	glVertex2f(1.0f, 1.0f);
	glTexCoord2f(0,ParticlesHeight);
	glVertex2f(-1.0f,1.0f);
	glEnd();
	glActiveTextureARB(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
	glActiveTextureARB(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
	glUseProgramObjectARB(0);

	Vector2f base = make_vector(rand() / float(RAND_MAX), rand() / float(RAND_MAX));
	Vector2f end = base;
	end += make_vector(SpawnSize / float(FieldWidth), SpawnSize / float(FieldHeight));
	end += make_vector(rand() / float(RAND_MAX) - SpawnSize / float(FieldWidth), rand() / float(RAND_MAX) - SpawnSize / float(FieldHeight));
	//spawn:
	{
		glUseProgramObjectARB(spawn_posvel_shader->handle);

		glActiveTextureARB(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, field[0].tex);
		glActiveTextureARB(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, spawn_tex);
		glBegin(GL_QUADS);
		glTexCoord2f(base.x, base.y);
		float c = 0.0; //no velocity variation, either!
		glColor4f(rand() / float(RAND_MAX) * c, rand() / float(RAND_MAX) * c, 0.0f, 0.0f);
		glVertex2f(spawn_box.min.x, spawn_box.min.y);
		glVertex2f(spawn_box.min.x, spawn_box.max.y);
		glColor4f(rand() / float(RAND_MAX) * c, rand() / float(RAND_MAX) * c, 0.0f, 0.0f);
		glTexCoord2f(end.x, end.y);
		glVertex2f(spawn_box.max.x, spawn_box.max.y);
		glVertex2f(spawn_box.max.x, spawn_box.min.y);
		glEnd();
		glActiveTextureARB(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, 0);
		glActiveTextureARB(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, 0);
		glUseProgramObjectARB(0);
	}

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, color_temp_fb);

	glUseProgramObjectARB(update_color_shader->handle);
	glActiveTextureARB(GL_TEXTURE3_ARB);
	glBindTexture(GL_TEXTURE_2D, cam_color);
	glActiveTextureARB(GL_TEXTURE2_ARB);
	glBindTexture(GL_TEXTURE_2D, field_color.tex);
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, color_tex);
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, posvel_tex);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex2f(-1.0f,-1.0f);
	glTexCoord2f(ParticlesWidth,0);
	glVertex2f(1.0f,-1.0f);
	glTexCoord2f(ParticlesWidth,ParticlesHeight);
	glVertex2f(1.0f, 1.0f);
	glTexCoord2f(0,ParticlesHeight);
	glVertex2f(-1.0f,1.0f);
	glEnd();

	glActiveTextureARB(GL_TEXTURE3_ARB);
	glBindTexture(GL_TEXTURE_2D, 0);
	glActiveTextureARB(GL_TEXTURE2_ARB);
	glBindTexture(GL_TEXTURE_2D, 0);
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
	glUseProgramObjectARB(0);

	//spawn color:
	glActiveTextureARB(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, field_color.tex /* testing... cam_color */);
	glActiveTextureARB(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, spawn_tex);
	glUseProgramObjectARB(spawn_color_shader->handle);
	glBegin(GL_QUADS);
	float c = 0.0; //no color variation, for now.
	glColor4f(rand() / float(RAND_MAX) * c, rand() / float(RAND_MAX) * c, rand() / float(RAND_MAX) * c, 1.0f);
	glTexCoord2f(base.x, base.y);
	glVertex2f(spawn_box.min.x, spawn_box.min.y);
	glVertex2f(spawn_box.min.x, spawn_box.max.y);
	glColor4f(rand() / float(RAND_MAX) * c, rand() / float(RAND_MAX) * c, rand() / float(RAND_MAX) * c, 1.0f);
	glTexCoord2f(end.x, end.y);
	glVertex2f(spawn_box.max.x, spawn_box.max.y);
	glVertex2f(spawn_box.max.x, spawn_box.min.y);
	glEnd();
	glUseProgramObjectARB(0);

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	glPopAttrib();
	swap(posvel_tex, posvel_temp_tex);
	swap(posvel_fb, posvel_temp_fb);
	swap(color_tex, color_temp_tex);
	swap(color_fb, color_temp_fb);

	spawn_point = (spawn_point + SpawnSize) % (ParticlesWidth * ParticlesHeight);

	Graphics::gl_errors("tick");
}
